/**
 * @author Micha Sobiecki grupa J1I2
 * 
 * Praca Domowa ZPR - przykad uycia kontenera std::map
 * 
 * zabezpieczenie przed powieleniem zasobw oparte na std::map
 */

#include <string>
#include <map>
#include <iostream>

/**
 * Klasa zajmuje si udostpnianiem zasobw w postaci tekstur graficznych.
 * Jest to jedynie szkielet, ktry w rzeczywistoci nie wczytuje tekstur,
 * jednak celem kodu jest pokazanie idei w jaki sposb mona prosto i szybko zrealizowa taki system.
 * W celu zacignicia tego kodu do dziaania z rzeczywistymi teksturami wystarczy zamieni kod funkcji load() 
 * dodajc np instrukcje z biblioteki DevIL oraz OpenGL, wykraczao to jednak poza ramy tego zadania.
 * Manager moe by rwnie uyty do zasobw dowolnego innego typu po niewielkich przerbkach.
 */
class TextureManager {
public:
	/// Struktura danych przechowujca potrzebne informacje o wczytanej teksturze.
	struct Texture {
		/// ID_ tekstury zwracany przez sterownik karty graficznej po mapowaniu w pamici
		unsigned int ID_;
		/// szeroko tekstury wyraona w tekselach
		unsigned int width_;
		/// wysoko tekstury wyraona w tekselach
		unsigned int height_;
		/// ilo bitw przypadajcych na jeden teksel
		unsigned int bpp_;
	};

private:
	/// Kontener asocjacyjny przechowujcy informacje o wczytanych teksturach.
	/// Jako klucz danej tekstury przechowywana jest cieka do pliku z ktrego zostaa wczytana,
	/// co zapewnia unikalno klucza.
	std::map<std::string, Texture> loaded_;

	// konstruktory zostay ukryte, poniewa Manager realizowany jest jako singleton
	TextureManager() {}
	TextureManager(const TextureManager& m) {}
public:
	/// metoda zwraca instancj singletonu TextureManager
	static TextureManager& get() {
		static TextureManager TM;
		return TM;
	}

	/**
	 * Zadaniem funkcji jest adowanie tekstury z pliku.
	 * W przypadku, gdy tekstura ju raz bya zaadowana,
	 * w celu oszczdnoci pamici karty graficznej,
	 * funkcja zwrci identyfikator tekstury zaadowanej wczeniej.
	 * @param name cieka do pliku w ktrym znajduje si tekstura
	 * @return identyfikator tekstury
	 * @see Texture::ID_
	 */
	unsigned int load(const std::string& name);

	/**
	 * Sprawdza czy zasb zosta ju zaadowany
	 * @param name cieka do pliku w ktrym znajduje si tekstura
	 * @return true jeli tekstura ju jest zaadowana
	 */
	bool isLoaded(const std::string& name);

	/**
	 * Zwalnia zasb (usuwa tekstur z mapy)
	 * W przypadku gdy tekstura nie istnieje nic nie robi.
	 * @param name cieka tekstury do usunicia
	 */
	void free(const std::string& name);

	/**
	 * @return liczba zaadowanych tekstur
	 */
	unsigned int numberOfResources();
};

unsigned int TextureManager::load(const std::string& name) {
	std::map<std::string, Texture>::iterator i;
	i = loaded_.find(name);
		
	if(i != loaded_.end())
		return i->second.ID_;	// tekstura juz raz zostaa wczytana

	// sekcja wczytywania obrazka, mona zrealizowa np poprzez bibliotek DevIL
	// tutaj ustawiane s przykadowe dane nie wczytywane z pliku.
	Texture tex;
	tex.width_ = 64;	// 64 texele szerokoci
	tex.height_ = 64;	// 64 texele wysokoci
	tex.bpp_ = 32;		// 32 bity na texel

	// nastpnie powinno nastpi mapowanie tekstury do pamici karty graficznej,
	// mapowanie takie zwraca (np. w OpenGL) identyfikator tekstury w postaci liczby cakowitej nieujemnej (GLuint)
	// identyfikatora mona nastpnie uy (glBind(<typ tekstury>, ID)) w celu okrelenia, e tekstura ma by rysowana na ekranie
	// dla uatwienia proces ten jest zastpiony zwyk inkrementacj identyfikatora
	// kada nowo wczytana tekstura dostanie ID o jeden wyszy ni poprzednia, pierwsza dostanie 0.
	static unsigned int nextID = 0;
	tex.ID_ = nextID;
	nextID++;

	// dodanie tekstury do mapy
	loaded_[name] = tex;

	return tex.ID_;
}

bool TextureManager::isLoaded(const std::string& name) {
	// count zwraca ilosc elementow o kluczu podanym jako parametr (1 lub 0)
	if(loaded_.count(name) > 0)
		return true;

	return false;
}

void TextureManager::free(const std::string& name) {
	std::map<std::string, Texture>::iterator i;
	i = loaded_.find(name);
		
	if(i == loaded_.end())
		return;

	// usuniecie elementu wskazywanego przez iterator
	loaded_.erase(i);
}

unsigned int TextureManager::numberOfResources() {
	// std::map::size() zwraca ilosc elementow w mapie.
	return static_cast<unsigned int>(loaded_.size());
}

int main() {
	TextureManager& TM = TextureManager::get();

	std::cout << "TM.numberOfResources() = " << TM.numberOfResources() << std::endl;
	std::cout << "TM.load(\"tex1.jpg\") = " << TM.load("tex1.jpg") << std::endl;
	std::cout << "TM.load(\"tex2.jpg\") = " << TM.load("tex2.jpg") << std::endl;
	std::cout << "TM.load(\"tex1.jpg\") = " << TM.load("tex1.jpg") << std::endl;
	std::cout << "TM.load(\"tex2.jpg\") = " << TM.load("tex2.jpg") << std::endl;
	std::cout << "TM.load(\"tex3.jpg\") = " << TM.load("tex3.jpg") << std::endl;
	std::cout << "TM.load(\"tex4.jpg\") = " << TM.load("tex4.jpg") << std::endl;
	std::cout << "TM.load(\"tex4.jpg\") = " << TM.load("tex4.jpg") << std::endl;
	std::cout << "TM.load(\"tex2.jpg\") = " << TM.load("tex2.jpg") << std::endl;
	std::cout << "TM.numberOfResources() = " << TM.numberOfResources() << std::endl;
	std::cout << "TM.free(\"tex1.jpg\") = " << std::endl;
	TM.free("tex1.jpg");
	std::cout << "TM.numberOfResources() = " << TM.numberOfResources() << std::endl;
	std::cout << "TM.load(\"tex1.jpg\") = " << TM.load("tex1.jpg") << std::endl;
	if(TM.isLoaded("tex1.jpg"))
		std::cout << "tex1 is loaded\n";
	if(TM.isLoaded("tex5.jpg"))
		std::cout << "tex5 is loaded\n";

	return 0;
}

// WYNIK DZIALANIA PROGRAMU:

//TM.numberOfResources() = 0
//TM.load("tex1.jpg") = 0
//TM.load("tex2.jpg") = 1
//TM.load("tex1.jpg") = 0
//TM.load("tex2.jpg") = 1
//TM.load("tex3.jpg") = 2
//TM.load("tex4.jpg") = 3
//TM.load("tex4.jpg") = 3
//TM.load("tex2.jpg") = 1
//TM.numberOfResources() = 4
//TM.free("tex1.jpg") =
//TM.numberOfResources() = 3
//TM.load("tex1.jpg") = 4
//tex1 is loaded
